#include <algorithm>
#include <vector>
#include <array>
#include <chrono>
#include <cstdio>
#include <cassert>

const size_t size = 1<<25;

std::chrono::high_resolution_clock::time_point timer_start;
typedef std::vector<int>::iterator iterator;

void start() {
    timer_start = std::chrono::high_resolution_clock::now();
}

void stop(const char* message) {
    float elapsedTime = std::chrono::duration_cast<std::chrono::duration<float>>(std::chrono::high_resolution_clock::now() - timer_start).count();
    printf("%s %f\n", message, elapsedTime);
    start();
}

void check_sorted(int* begin, int* end) {
    for (int *p=begin+1; p < end; p++) {
        assert(*(p-1) <= *p);
    }
}

template<typename T>
inline void bubblesort(T begin, T end) {
    while (end > begin) {
        end -= 1;
        for (T p=begin; p<end; p++) {
            if (*p > *(p+1)) {
                std::swap(*p, *(p+1));
            }
        }
    }
}

inline void make_pivot(iterator begin, iterator end) {
    // std::swap(begin[(end - begin)/2], end[-1]);
    // return;
    iterator a = begin;
    iterator c = begin + (end - begin) / 2;
    iterator b = end - 1;
    if (*c < *a) std::swap(a, c);
    if (*c < *b) std::swap(*c, *b);
    else if (*a > *b) std::swap(*a, *b);
    assert(*a <= *b);
    assert(*b <= *c);
}

inline iterator partition_twosided(iterator begin, iterator end, int pivot) {
    iterator left = begin, right = end - 1;
    while (true) {
        while (*left < pivot) {
            left += 1;
        }
        while (*right >= pivot) {
            if (right == left) {
                return left;
            }
            right -= 1;
        }
        std::swap(*left, *right);
    }
}

inline iterator partition_onepass(iterator begin, iterator end, int pivot) {
    for (iterator mid = begin; mid < end; mid++) {
        if (*begin < pivot) {
            std::swap(*mid, *begin);
            begin += 1;
        }
    }
    return begin;
}

size_t count_calls = 0;
void quicksort(iterator begin, iterator end) {
    count_calls += 1;
    if (end - begin <= 1) {
        return;
    }
    make_pivot(begin, end);
    end -= 1;
    int pivot = *end;
    iterator mid;
    if (0) {
        mid = partition_twosided(begin, end, pivot);
    } else if (1) {
        mid = partition_onepass(begin, end, pivot);
    } else {
        mid = std::partition(begin, end, [pivot](int i){ return i < pivot; });
    }
    std::swap(*mid, *end);
    quicksort(begin, mid);
    quicksort(mid + 1, end);
}

int main() {
    std::vector<int> data;
    for (size_t i=0; i<size; i++) {
        data.push_back((i%7)*(i%1234)*(i%4321));
    }
    start();
    if (0) {
        std::sort(data.begin(), data.end());
        stop("std::sort");
    } else {
        quicksort(data.begin(), data.end());
        stop("recursive quicksort");
    }
    printf("%lu calls on %lu elements\n", count_calls, size);
    if (size < 100) {
        for (size_t i=0; i<size; i++) {
            printf("%d ", data[i]);
        }
        printf("\n");
    } else {
        check_sorted(&data[0], &data[size - 1]);
    }
}